EXPORTS
Init_narray
na_sizeof
na_make_object
na_make_scalar
na_make_empty
na_get_typecode
na_clear_data
na_clone
na_fill
na_copy_nary
na_to_array
na_ary_to_nary
na_object_type
na_cast_object
na_cast_unless_narray
na_cast_unless_array
na_upcast_object
na_dup_w_type
na_change_type
na_upcast_type
na_to_narray
na_aset
na_aref
na_slice
na_count_true
na_count_false
na_aref_mask
na_aset_mask
